import java.net.*;
import java.io.*;
import java.util.*;

class responseData {
    public int statusCode;
    public String statusText;
    public String requestType;
    public String resourceName;
    public String contentType;
    public int contentLength;
}

    class ClientHandler extends Thread {
    
    Socket socket; //local pointer to socket being processed
    Semaphore threadLock;
    ClientResponder responderThread;
    Semaphore activeSem;
    MIMETypes MIMEmap;
    boolean connection;
    boolean activeStatus;
    
    ClientHandler(ClientResponder thread, Semaphore activeThreadCount)
    {
        threadLock = new Semaphore(0);
        activeStatus=false;
        activeSem = activeThreadCount;
        responderThread=thread;
        MIMEmap = new MIMETypes();
    }
    
    public void run() //thread's primary function
    {
        int tempInt;
        int waitCount=0;
        char[] temp = new char[1000];
        String tempString = new String();
        String request;
        BufferedReader in;
        String dataBuffer;
        responseData header;

        System.out.println("    - clientHandler thread running");
        
        while(true)
        {
            try
            {
                threadLock.acquire(); //try to acquire lock.  If sucessful, new connection is ready to be processed

                while (connection==true)
                {
                    request = "    "; //initalize request string
                    dataBuffer=""; //initalize dataBuffer to be empty with size 0;
                    waitCount=0;
                    in = new BufferedReader(new InputStreamReader(socket.getInputStream()));
                    //initialize object responsible for reading data from socket
                    
                    while (connection==true && !(request.charAt(request.length()-1)==(char)10 && request.charAt(request.length()-2)==(char)13 && request.charAt(request.length()-3)==(char)10 && request.charAt(request.length()-4)==(char)13))
                    //continue reading out of socket until two CRLFs are encountered or connection times out
                    {
                        if (in.ready())
                        {
                            tempInt = in.read(temp,0,1000); //try to read 1000 bytes
                            request += tempString.copyValueOf(temp,0,tempInt); //add bytes to request String
                        }
                        else
                            sleep(200);
                        
                        if (this.responderThread.busy==false)
                            waitCount++;
                        
                        //System.out.println(this.responderThread.busy);
                        
                        if (waitCount>0 && this.responderThread.busy)
                            waitCount=0;
                        
                        if (waitCount>5) //if 1 seconds has elapsed with no new requests or transmissions, close connection
                        {
                            System.out.println("        > Client connection timed out.");
                            connection=false;
                        }
                    }
                    
                    //System.out.println(request);
                    
                    if (request.indexOf("Connection: close")>=0)  //check if server indicated to close connection
                    {
                        header = requestParse(request); //pass request String to seperate function for parsing
                        responderThread.newRequest(header); //notify responderThread about new request
                        connection=false;
                        System.out.println("        > Client closed connection");
                    }
                    
                    if (connection==true)
                    {
                        System.out.println("        > Request received.  Notifying responderThread.");
                        header = requestParse(request); //pass request String to seperate function for parsing
                        responderThread.newRequest(header); //notify responderThread about new request
                    }
                }
                
                System.out.println("    + All requests satisfied.  Preparing to close socket and free thread.");
                System.out.println("        > Checking all responses have been sent before closing socket");
                this.responderThread.requestComplete.acquire(); //check that all responses have been sent
                System.out.println("        > All responses have been sent! Closing Socket. Thread returning to pool.");
                socket.close(); //requests completed, close socket
                activeStatus=false;
                activeSem.release();   //request completed, inform threadHandler thread is ready to process new connection
            }
            catch (Exception e) //catch any exceptions
            {
                System.err.println(e);
            }
        }
    }
    
    public void giveSocket(Socket s)
    {
        this.activeStatus=true;
        this.connection=true;  
        this.socket=s;
        responderThread.giveSocket(s);
        this.threadLock.release();
    }
    
    public boolean busy()
    {
        return activeStatus;
    }
    
    public responseData requestParse(String request)
    {       
        String temp;
        
        responseData header =  new responseData(); //initialize new responseData object to retain data
        StringTokenizer tokens = new StringTokenizer(request, " "); //tokenize request for easier processing
        
        temp=tokens.nextToken(); //get first token
        
        if (temp.compareTo("GET")==0)
            header.requestType="GET";
        
        if (temp.compareTo("HEAD")==0)
            header.requestType="HEAD";
        
        header.resourceName = (tokens.nextToken()).substring(1);
        header.contentType = MIMEmap.getMIMEType(header.resourceName.substring(header.resourceName.indexOf('.')));
        
        File inputFile = new File(header.resourceName);
        
        if (inputFile.exists()) //if file was found, set status code = 200
        {
            header.statusCode=200;
            header.statusText="OK";
            header.contentLength=(int)inputFile.length(); //initilize content length based on file size
        }
        else //if file was not found, set status code = 404
        {
            header.statusCode=404;
            header.statusText="Not Found";
        }
        return header;
    }
}